# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""
Spock is a framework that helps manage complex parameter configurations for Python applications

Please refer to the documentation provided in the README.md
"""
from spock.addons.tune.config import (
    ChoiceHyperParameter,
    OptunaTunerConfig,
    RangeHyperParameter,
    spockTuner,
)

__all__ = [
    "builder",
    "config",
    "spockTuner",
    "RangeHyperParameter",
    "ChoiceHyperParameter",
    "OptunaTunerConfig",
]
